/**
Copyright (c) 2000 Brian Nenninger

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**/

#import "LifeSaver.h"
#include <CoreFoundation/CoreFoundation.h>

static int MAX_SPEED = 8;
static NSString *SUITE_NAME = @"com.dozingcat.LifeSaver";

@implementation LifeSaver

- (id)initWithFrame:(NSRect)frameRect isPreview:(BOOL)preview
{
    self = [super initWithFrame:frameRect isPreview:preview];
    if (self) {
	NSOpenGLPixelFormatAttribute attribs[] =
        {	NSOpenGLPFAAccelerated,
		NSOpenGLPFAColorSize, 16,
		NSOpenGLPFADepthSize, 16,
		NSOpenGLPFAMinimumPolicy,
		NSOpenGLPFAClosestPolicy,
                0
        };
	
	NSOpenGLPixelFormat *format = [[[NSOpenGLPixelFormat alloc] initWithAttributes:attribs] autorelease];
        _view = [[[NSOpenGLView alloc] initWithFrame:NSZeroRect pixelFormat:format] autorelease];
        scene = nil;
        configureSheet = nil;
        
        CFPreferencesAddSuitePreferencesToApp(kCFPreferencesCurrentApplication, 
                                              (CFStringRef)SUITE_NAME);

        [self addSubview:_view];
        [self updateParameters];
    }

    return self;
}

-(BOOL)writePrefsDictionary:(NSDictionary *)dict {
  //NSLog(@"Writing dictionary %@", dict);
  //return [dict writeToFile:[@"~/Library/Preferences/LifeSaver.plist" stringByExpandingTildeInPath] atomically:YES];
  CFPreferencesSetAppValue((CFStringRef)@"prefs", (CFDictionaryRef)dict, (CFStringRef)SUITE_NAME);
  return YES;
}

-(NSDictionary *)readPrefsDictionary {
  return (NSDictionary *)CFPreferencesCopyAppValue((CFStringRef)@"prefs", (CFStringRef)SUITE_NAME);;
  //return [NSDictionary dictionaryWithContentsOfFile:[@"~/Library/Preferences/LifeSaver.plist" stringByExpandingTildeInPath]];
}

-(BOOL)saveParameters {
  NSMutableDictionary *dict = [[[NSMutableDictionary alloc] init] autorelease];
  [dict setObject:[NSNumber numberWithInt:xcells] forKey:@"GridX"];
  [dict setObject:[NSNumber numberWithInt:ycells] forKey:@"GridY"];
  [dict setObject:[NSNumber numberWithInt:zcells] forKey:@"GridZ"];
  [dict setObject:[NSNumber numberWithInt:useAllScreens] forKey:@"UseAllScreens"];
  [dict setObject:[NSNumber numberWithInt:sphereDetail] forKey:@"SphereDetail"];
  [dict setObject:[NSNumber numberWithInt:simspeed] forKey:@"SimSpeed"];
  return [self writePrefsDictionary:dict];
}

-(void)updateParameters {
  int dims[3];
  // read from defaults
  NSDictionary *configDict = [self readPrefsDictionary];
  // these will all return 0 if there is no preferences plist
  xcells = [[configDict objectForKey:@"GridX"] intValue];
  ycells = [[configDict objectForKey:@"GridY"] intValue];
  zcells = [[configDict objectForKey:@"GridZ"] intValue];
  useAllScreens = [[configDict objectForKey:@"UseAllScreens"] intValue];
  simspeed      = [[configDict objectForKey:@"SimSpeed"] intValue];
  sphereDetail  = [[configDict objectForKey:@"SphereDetail"] intValue];
  // assign default values if not present in properties
  if (zcells<=0) zcells = 25;
  if (ycells<=0) ycells = 25;
  if (xcells<=2) xcells = 0;
  if (sphereDetail<=0) sphereDetail = 8;
  
  // set ivars, using default values if NSUserDefaults didn't exist
  dims[0] = ycells;
  dims[1] = zcells;
  dims[2] = xcells;
  numDimensions = (dims[2] > 1) ? 3 : 2;
  
  // compute # of transition and fixed frames based on simspeed
  if (simspeed>=MAX_SPEED) {
    // max possible speed
    transitionFrames = fixedFrames = 1;
  }
  else {
    // higher speed gives fewer frames
    transitionFrames = MAX_SPEED - simspeed;
    fixedFrames = 2*transitionFrames;
  }
    
  // these parameters are fixed for now
  // y and z range from -5 to +5, base radius is 4
  maxYZCoord = 5.0; 
  maxRotateSpeed = 10;
  baseRadius = 4.0/dims[1];
  rotateFactor = 0.006;
  rotateSpeed = 2;
  gridRepetitionLimit   = 20;
  gridRepetitionCounter = 0;
  
  spaceBetweenCells = (2*maxYZCoord)/(ycells-1);
  maxXCoord = (xcells>1) ? ((xcells-1)*spaceBetweenCells/2) : 0;
    
  [scene autorelease];
  scene = [[OCLifeHypergrid alloc] initWithNumberOfDimensions:numDimensions sizes:dims];
}

- (NSTimeInterval)animationTimeInterval
{
    return 0.01;
}

- (void)setFrameSize:(NSSize)newSize
{
    [super setFrameSize:newSize];
    [_view setFrameSize:newSize];
    _initedGL = NO;

}

- (void)animateOneFrame
{
  if (![self shouldDraw]) {
    return;
  }
    [[_view openGLContext] makeCurrentContext];
    if (!_initedGL) {
      [self initGL];
      _initedGL = YES;
    }
  
  angle += (rotateFactor*rotateSpeed);
  if (angle>6.283) angle -= 6.283;
  if (angle<0) angle += 6.283;
  
  if (!isPaused) {
    // increment counter and possibly switch in/out of transition mode
    counter++;
    if (inTransition && counter>transitionFrames) {
      // switching to static view, update life grid
      [scene iterate];
      counter = 0;
      inTransition = NO;
      // if we're in a repetition, increment repetition counter and restart if it hits limit
      if (gridRepetitionCounter>0) {
        if (++gridRepetitionCounter >= gridRepetitionLimit) {
          [scene randomizeGrid];
          gridRepetitionCounter = 0;
        }
      }
      else {
        // check for repetition, if true start repetition counter
        if ([scene matchesGridInHistory]) {
          gridRepetitionCounter = 1;
        }
      }
    }
    else if (!inTransition && counter>fixedFrames) {
      // beginning transition display, don't update grid
      counter = 0;
      inTransition = YES;
    }
    
  }

        
    [self draw];
}

-(void)increaseRotation {
  if (rotateSpeed<maxRotateSpeed) rotateSpeed++;
}

-(void)decreaseRotation {
  if (rotateSpeed>-maxRotateSpeed) rotateSpeed--;
}

-(void)keyDown:(NSEvent *)event {
  NSString *chars = [event characters];
  if ([chars isEqualToString:@"r"]) [scene randomizeGrid];
  else if ([chars isEqualToString:@"p"]) isPaused = !isPaused;
  else if ([chars isEqualToString:@"s"]) [self increaseRotation];
  else if ([chars isEqualToString:@"a"]) [self decreaseRotation];
  else [super keyDown:event];
}
/*
-(void)mouseUp:(NSEvent *)event {

}
*/
- (BOOL)hasConfigureSheet { return YES; }
- (NSWindow*)configureSheet {
  if (!configureSheet) {
    [NSBundle loadNibNamed:@"ConfigureSheet" owner:self];
  }
  // set control values
  [size2dField setIntValue:ycells];
  [size3dField setIntValue:xcells];
  [detailSlider setIntValue:sphereDetail];
  [speedSlider setIntValue:simspeed];
  [allScreensCheckbox setState:useAllScreens];
  return configureSheet;
}


/************************************************************
Methods for configuration sheet
************************************************************/

-(void)saveSettings:(id)sender {
  zcells = ycells = [size2dField intValue];
  xcells = [size3dField intValue];
  sphereDetail = [detailSlider intValue];
  simspeed = [speedSlider intValue];
  useAllScreens = [allScreensCheckbox intValue]; // assuming ints convert to BOOLs nicely

  // save and reinitialize grid
  [self saveParameters];
  [self updateParameters];
  [NSApp endSheet:configureSheet];
}

-(void)cancelSettings:(id)sender {
    [NSApp endSheet:configureSheet];
}

/************************************************************
Drawing methods
************************************************************/
-(void)initGL {
    static float ambient_light[] = {0.5, 0.5, 0.5, 1.0};
    angle = 0.0;
    dist  = 10.0;
    sphereQObj = NULL;
    
    glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
    glEnable(GL_COLOR_MATERIAL);    
    glShadeModel(GL_SMOOTH);
    glEnable(GL_DEPTH_TEST);
    
    glLightfv(GL_LIGHT0, GL_AMBIENT, ambient_light);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);

    
    glMatrixMode(GL_PROJECTION);
    gluPerspective( /* field of view in degrees */ 60.0,
    /* aspect ratio */ 4.0/3.0,
    /* Z near */ 0.1, /* Z far */ 20);

    glMatrixMode(GL_MODELVIEW);    
        
    glFinish();
}

-(void)draw {
    NSEnumerator *gridenum;
    id coords;
    glDrawBuffer(GL_BACK);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    if (!sphereQObj) sphereQObj = gluNewQuadric();

    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    
    gluLookAt(dist*cos(angle), 0, dist*sin(angle),
               -cos(angle), 0, -sin(angle),
               0, 1, 0);
    
    gridenum = [scene coordinateEnumerator];
    while (coords=[gridenum nextObject]) {
      if (!inTransition) {
        // static view, spheres are full size, either blue or red
        if ([scene valueAtCoordinates:coords]) {
          if ([scene cellWillSurviveAtCoordinates:coords]) {
            // blue for alive
            glColor3f(0, 0, 1);
          }
          else {
            // red for about to die
            glColor3f(1, 0, 0);
          }
          [self drawCellAtCoordinates:coords scaledBy:1.0];
        }
      }
      else {
        // dynamic view, red spheres shrink, green ones grow
        if ([scene valueAtCoordinates:coords]) {
          if ([scene cellWillSurviveAtCoordinates:coords]) {
            // full size, blue
            glColor3f(0, 0, 1);
            [self drawCellAtCoordinates:coords scaledBy:1.0];
          }
          else {
            // shrinking, red
            double scale = (1.0-((double)counter)/transitionFrames);
            glColor3f(1, 0, 0);
            [self drawCellAtCoordinates:coords scaledBy:scale];
          }
        }
        else {
          if ([scene cellWillBeBornAtCoordinates:coords]) {
            // growing, green
            double scale = (((double)counter)/transitionFrames);
            glColor3f(0, 1, 0);
            [self drawCellAtCoordinates:coords scaledBy:scale];
          }
        }
      }
    }
        
    glFinish();
}

-(void)drawCellAtCoordinates:(OCHypergridCoordinates *)coords scaledBy:(double)scale {
  int zc = [coords valueAtIndex:1];
  int yc = [coords valueAtIndex:0];
  double y = -maxYZCoord + (yc*spaceBetweenCells);
  double z = -maxYZCoord + (zc*spaceBetweenCells);
  double x = 0;
  if (numDimensions>2) {
    int xc = [coords valueAtIndex:2];
    x = -maxXCoord + (xc*spaceBetweenCells);
  }
  glPushMatrix();
  glTranslatef(x, y, z);
  gluSphere(sphereQObj, baseRadius*scale, sphereDetail, sphereDetail);
  glPopMatrix();          
}

-(BOOL)shouldDraw {
  // If flag is set, don't draw on non-main screen.
  if (useAllScreens) return YES;
  else return ([[_view window] screen] == [NSScreen mainScreen]);
}

-(void)dealloc {
  [scene release];
  [configureSheet release];
  [super dealloc];
}


@end
