/**
Copyright (c) 2000 Brian Nenninger

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**/

#import "OCLifeHypergrid.h"


@implementation OCLifeHypergrid

-(id)init {
  self = [super init];
  if (self) {
    historySize = 4;
    previousGrids = [[NSMutableArray alloc] initWithCapacity:historySize];
  }
  return self;
}

-(id)initWithNumberOfDimensions:(int)num sizes:(int *)sizes {
  self = [self init];
  if (self) {
    lifegrid = [[OCHypergrid alloc] initWithNumberOfDimensions:num sizes:sizes];
    neighborgrid = [[OCHypergrid alloc] initWithNumberOfDimensions:num sizes:sizes];
    [self randomizeGrid];
  }
  return self;
}

-(void)randomizeGrid {
  // fill lifegrid with random pattern
  int p = ([lifegrid numberOfDimensions]>2) ? 0x7000 : 0x1000;
  [previousGrids removeAllObjects];
  srand(time(NULL));
  {
    NSEnumerator *gridenum = [lifegrid coordinateEnumerator];
    id coords;
    while (coords=[gridenum nextObject]) {
      int state = ((rand() & p)==p);
      [lifegrid setValue:state atCoordinates:coords];
    }
  }
  [self createNeighborGrid];
}

-(void)createNeighborGrid {
  NSEnumerator *gridenum;
  id coords;
  gridenum = [lifegrid coordinateEnumerator];
  while (coords=[gridenum nextObject]) {
    int n = [lifegrid nonzeroNeighborsOfCoordinates:coords];
    [neighborgrid setValue:n atCoordinates:coords];
  }
}

-(BOOL)iterate {
  // rotate x-z
  NSEnumerator *gridenum;
  id coords;
  BOOL changed = NO;
  // save grid to history
  [self recordGridInHistory];
  // apply rules (this assumes neighborgrid is already set up)
  gridenum = [lifegrid coordinateEnumerator];
  while (coords=[gridenum nextObject]) {
    int val = [lifegrid valueAtCoordinates:coords];
    int neighbors = [neighborgrid valueAtCoordinates:coords];
    if (val) {
      // existing cell must have 2 or 3 neighbors to live
      if (neighbors<2 || neighbors>3) {
        [lifegrid setValue:0 atCoordinates:coords];
        changed = YES;
      }
    } 
    else {
      // empty cell becomes active with 3 neighbors
      if (neighbors==3) {
        [lifegrid setValue:1 atCoordinates:coords];
        changed = YES;
      }
    }
  }
  [self createNeighborGrid];
  return changed;
}

-(BOOL)cellWillSurviveAtCoordinates:(OCHypergridCoordinates *)coords {
  int neighbors = [neighborgrid valueAtCoordinates:coords];
  return (neighbors==2 || neighbors==3);
}

-(BOOL)cellWillBeBornAtCoordinates:(OCHypergridCoordinates *)coords {
  int neighbors = [neighborgrid valueAtCoordinates:coords];
  return (neighbors==3);
}

-(void)recordGridInHistory {
  if ([previousGrids count]==historySize) [previousGrids removeObjectAtIndex:(historySize-1)];
  [previousGrids insertObject:[[[self lifeHypergrid] copy] autorelease] atIndex:0];
}

-(BOOL)matchesGridInHistory {
  NSEnumerator *gridEnum = [previousGrids objectEnumerator];
  id grid;
  while (grid = [gridEnum nextObject]) {
    if ([[self lifeHypergrid] isEqualToHypergrid:grid]) return YES;
  }
  return NO;
}

-(OCHypergrid *)lifeHypergrid {
  return lifegrid;
}

-(OCHypergrid *)neighborHypergrid {
  return neighborgrid;
}

-(gridvalue_t)valueAtCoordinates:(OCHypergridCoordinates *)coords {
  return [lifegrid valueAtCoordinates:coords];
}

-(gridvalue_t)neighborsOfCoordinates:(OCHypergridCoordinates *)coords {	
  return [neighborgrid valueAtCoordinates:coords];
}

-(NSEnumerator *)coordinateEnumerator {
  return [lifegrid coordinateEnumerator];
}

-(void)dealloc {
  [lifegrid release];
  [neighborgrid release];
  [previousGrids release];
}



@end
