/* -*- Mode: C; tab-width: 4 -*- */
/* superquadrics --- 3D mathematical shapes */

/*-
 * Permission to use, copy, modify, and sq_distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This file is provided AS IS with no warranties of any kind.  The author
 * shall have no liability with respect to the infringement of copyrights,
 * trade secrets or any patents by this file or any part thereof.  In no
 * event will the author be liable for any lost revenue or profits or
 * other special, indirect and consequential damages.
 *
 * Superquadrics were invented by Dr. Alan Barr of Caltech University.
 * They were first published in "Computer Graphics and Applications",
 * volume 1, number 1, 1981, in the article "Superquadrics and Angle-
 * Preserving Transformations."  Dr. Barr based the Superquadrics on
 * Piet Hein's "super ellipses."  Super ellipses are like 2D ellipses,
 * except that the formula includes an exponent, raising its X and Y
 * values to a (fractional) power, and allowing them to gradually
 * change from round to square edges.  Superquadrics extend this
 * idea into 3 dimensions, using two exponents to modify a
 * quadric surface in a similar fashion.
 *
 * Revision History:
 * 30-Mar-97: Turned into a module for xlockmore 4.02 alpha.  The code
 *    is almost unrecognizable now from the first revision, except for
 *    a few remaining two-letter variable names.  I still don't have
 *    the normal vectors working right (I wrote the buggy normal vector
 *    code myself, can you tell?)
 * 07-Jan-97: A legend reborn; Superquadrics make an appearance as a
 *    real OpenGL program written in C.  I can even render them with
 *    proper lighting and specular highlights.  Gee, they look almost
 *    as good now as the original color plates of them that my uncle
 *    showed me as a child in 1981.  I don't know what computer hardware
 *    he was using at the time, but it's taken a couple decades for the
 *    PC clone hardware to catch up to it.
 * 05-Jan-97: After almost a decade, Superquadrics had almost faded away
 *    into the myths and folklore of all the things my brother and I played
 *    with on computers when we were kids with too much time on our hands.
 *    I had since gotten involved in Unix, OpenGL, and other things.
 *    A sudden flash of inspiration caused me to dig out the old Pascal
 *    source code, run it through p2c, and start ripping away the old
 *    sq_wireframe rendering code, to be replaced by OpenGL.
 * Late 1989 or early 1990:  Around this time I did the Turbo Pascal
 *    port of the Superquadrics.  Unfortunately, many of the original variable
 *    names remained the same from the C= 64 original.  This was unfortunate
 *    because BASIC on the c64 only allowed 2-letter, global variable names.
 *    But the speed improvement over BASIC was very impressive at the time.
 * Thanksgiving, 1987: Written.  My uncle Al, who invented Superquadrics some
 *    years earlier, came to visit us.  I was a high school kid at the time,
 *    with nothing more than a Commodore 64.  Somehow we wrote this program,
 *    (he did the math obviously, I just coded it into BASIC for the c64).
 *    Yeah, 320x200 resolution, colorless white sq_wireframe, and half an hour
 *    rendering time per superquadric.  PLOT x,y.  THOSE were the days.
 *    In the following years I would port Superquadrics to AppleBASIC,
 *    AmigaBASIC, and then Turbo Pascal for IBM clones.  5 minutes on a 286!
 *    Talk about fast rendering!  But these days, when my Pentium 166 runs
 *    the same program, the superquadric will already be waiting on the
 *    screen before my monitor can change frequency from text to graphics
 *    mode.  Can't time the number of minutes that way!  Darn ;)
 *
 * Ed Mackey
 */

/*-
 * due to a Bug/feature in VMS X11/Intrinsic.h has to be placed before xlock.
 * otherwise caddr_t is not defined correctly
 */

# define PROGCLASS					"Superquadrics"
# define HACK_INIT					init_superquadrics
# define HACK_DRAW					draw_superquadrics
# define DEFAULTS	"*delay:		100     \n"			\
					"*count:		25      \n"			\
					"*cycles:		40      \n"			\
					"*sq_wireframe:	False	\n"
/*-
 * Note for low-CPU-speed machines:  If your frame rate is so low that
 * attempts at animation appear futile, try using "-cycles 1", which puts
 * Superquadrics into kind of a slide-show mode.  It will still use up
 * all of your CPU power, but it may look nicer.
 */

#define DEF_SPINSPEED  "5.0"

/* static XrmOptionDescRec opts[] =
{
  {"-spinspeed", ".superquadrics.spinspeed", XrmoptionSepArg, (caddr_t) NULL}
};
static argtype vars[] =
{
  {(caddr_t *) & spinspeed, "spinspeed", "Spinspeed", DEF_SPINSPEED, t_Float}
};
static OptionStruct desc[] =
{
	{"-spinspeed num", "speed of rotation, in degrees per frame"}
};

ModeSpecOpt superquadrics_opts =
{sizeof opts / sizeof opts[0], opts, sizeof vars / sizeof vars[0], vars, desc};

#ifdef USE_MODULES
ModStruct   superquadrics_description =
{"superquadrics", "init_superquadrics", "draw_superquadrics", "release_superquadrics",
 "refresh_superquadrics", "init_superquadrics", NULL, &superquadrics_opts,
 1000, 25, 40, 1, 4, 1.0, "",
 "Shows 3D mathematical shapes", 0, NULL};

#endif
*/

#include <OpenGL/glu.h>
#include <bsd/stdlib.h>
#include <bsd/math.h>

#define MaxRes          50
#define MinRes          5

typedef double dimi[MaxRes + 1];

#define Bool int
#define LRAND random
#define MAXRAND RAND_MAX

static void reshape_superquadrics(int x, int y);

typedef struct {
	double      xExponent, yExponent;
	GLfloat     r[4], g[4], b[4];
	long        Mode;
	int         sq_rotx, sq_rotz;
} state;

	int         sq_dist, sq_wireframe, sq_flatshade, shownorms, maxcount, maxwait;
	int         sq_counter, viewcount, viewwait, sq_mono;
	GLfloat     curmat[4][4], sq_rotx, sq_roty, sq_rotz, spinspeed;
	/* In dimi: the first letter stands for cosine/sine, the second
	 *          stands for North, South, East, or West.  I think.
	 */
	dimi        cs, se, sw, sn, ss, ce, cw, cn, Prevxx, Prevyy, Prevzz,
	            Prevxn, Prevyn, Prevzn;
	double      xExponent, yExponent, Mode;
	int         resolution;
	state       now, later;

#define CLIP_NORMALS 10000.0

static int
myrand(int range)
{
	return ((int) (((float) range) * LRAND() / (MAXRAND)));
}

static float
myrandreal(void)
{
	return (LRAND() / (MAXRAND));
}

/* Some old, old, OLD code follows.  Ahh this takes me back..... */

/* Output from p2c, the Pascal-to-C translator */
/* From input file "squad.pas" */

static double
XtoY(double x, double y)
{
	double      z, a;

	/* This is NOT your typical raise-X-to-the-Y-power function.  Do not attempt
	 * to replace this with a standard exponent function.  If you must, just
	 * replace the "a = exp(y * log(z));" line with something faster.
	 */

	z = fabs(x);
	if (z < 1e-20) {
		a = 0.0;
		return a;
	}
	a = exp(y * log(z));
	if (a > CLIP_NORMALS)
		a = CLIP_NORMALS;
	if (x < 0)
		a = -a;
	return a;
}


static double
Sine(double x, double e)
{
	/* This is just the sine wave raised to the exponent.  BUT, you can't
	 * raise negative numbers to fractional exponents.  So we have a special
	 * XtoY routune which handles it in a way useful to superquadrics.
	 */

	return (XtoY(sin(x), e));
}


static double
Cosine(double x, double e)
{
	return (XtoY(cos(x), e));
}


static void
MakeUpStuff(int allstuff)
{
	static int  pats[4][4] =
	{
		{0, 0, 0, 0},
		{0, 1, 0, 1},
		{0, 0, 1, 1},
		{0, 1, 1, 0}
	};

	int         dostuff;
	int         t, pat;
	GLfloat     r, g, b, r2, g2, b2;

	/* randomize it. */

	if (maxcount < 2)
		allstuff = 1;
	dostuff = allstuff * 15;
	if (!dostuff) {
		dostuff = myrand(3) + 1;
		if (myrand(2) || (dostuff & 1))
			dostuff |= 4;
		if (myrand(2))
			dostuff |= 8;
	}
	if (dostuff & 1) {
		later.xExponent = (((long) floor(myrandreal() * 250 + 0.5)) / 100.0) + 0.1;
		later.yExponent = (((long) floor(myrandreal() * 250 + 0.5)) / 100.0) + 0.1;

		/* Increase the 2.0 .. 2.5 range to 2.0 .. 3.0 */
		if (later.xExponent > 2.0)
			later.xExponent = (later.xExponent * 2.0) - 2.0;
		if (later.yExponent > 2.0)
			later.yExponent = (later.yExponent * 2.0) - 2.0;
	}
	if (dostuff & 2) {
		do {
			later.Mode = myrand(3L) + 1;
		} while (!allstuff && (later.Mode == now.Mode));
		/* On init: make sure it can stay in mode 1 if it feels like it. */
	}
	if (dostuff & 4) {
		if (sq_mono) {
			if (sq_wireframe) {
				b = g = r = 1.0;
				b2 = g2 = r2 = 1.0;
			} else {
				b = g = r = (GLfloat) (140 + myrand(100)) / 255.0;
				b2 = g2 = r2 = ((r > 0.69) ? (1.0 - r) : r);
			}
		} else {
			r = (GLfloat) (40 + myrand(200)) / 255.0;
			g = (GLfloat) (40 + myrand(200)) / 255.0;
			b = (GLfloat) (40 + myrand(200)) / 255.0;

			r2 = ((myrand(4) && ((r < 0.31) || (r > 0.69))) ? (1.0 - r) : r);
			g2 = ((myrand(4) && ((g < 0.31) || (g > 0.69))) ? (1.0 - g) : g);
			b2 = ((myrand(4) && ((b < 0.31) || (b > 0.69))) ? (1.0 - b) : b);
		}

		pat = myrand(4);
		for (t = 0; t < 4; ++t) {
			later.r[t] = pats[pat][t] ? r : r2;
			later.g[t] = pats[pat][t] ? g : g2;
			later.b[t] = pats[pat][t] ? b : b2;
		}
	}
	if (dostuff & 8) {
		later.sq_rotx = myrand(360) - 180;
		later.sq_rotz = myrand(160) - 80;
	}
}

static void
inputs()
{
	int         iv;
	double      u, v, mode3, cn3, inverter2, flatu, flatv;

	if (Mode < 1.000001) {
		mode3 = 1.0;
		cn3 = 0.0;
		inverter2 = 1.0;
	} else if (Mode < 2.000001) {
		mode3 = 1.0;
		cn3 = (Mode - 1.0) * 1.5;
		inverter2 = (Mode - 1.0) * -2.0 + 1.0;
	} else {
		mode3 = (Mode - 1.0);
		cn3 = (Mode - 2.0) / 2.0 + 1.5;
		inverter2 = -1.0;
	}

	if (sq_flatshade) {
		flatu = M_PI / (resolution - 1);
		flatv = mode3 * M_PI / ((resolution - 1) * 2);
	} else {
		flatu = flatv = 0.0;
	}

	/* (void) printf("Calculating....\n"); */
	for (iv = 1; iv <= resolution; iv++) {

		/* u ranges from PI down to -PI */
		u = (1 - iv) * 2 * M_PI / (resolution - 1) + M_PI;

		/* v ranges from PI/2 down to -PI/2 */
		v = (1 - iv) * mode3 * M_PI / (resolution - 1) + M_PI * (mode3 / 2.0);

		/* Use of xExponent */
		se[iv] = Sine(u, xExponent);
		ce[iv] = Cosine(u, xExponent);
		sn[iv] = Sine(v, yExponent);
		cn[iv] = Cosine(v, yExponent) * inverter2 + cn3;

		/* Normal vector computations only */
		sw[iv] = Sine(u + flatu, 2 - xExponent);
		cw[iv] = Cosine(u + flatu, 2 - xExponent);
		ss[iv] = Sine(v + flatv, 2 - yExponent) * inverter2;
		cs[iv] = Cosine(v + flatv, 2 - yExponent);
	}			/* next */

	/* Now fix up the endpoints */
	se[resolution] = se[1];
	ce[resolution] = ce[1];

	if (Mode > 2.999999) {
		sn[resolution] = sn[1];
		cn[resolution] = cn[1];
	}
}


static void
DoneScale()
{
	double      xx, yy, zz, xp = 0, yp = 0, zp = 0, xn, yn, zn, xnp = 0,
	            ynp = 0, znp = 0;
	int         ih, iv;

	/* Hey don't knock my 2-letter variable names.  Simon's BASIC rules, man! ;-> */
	/* Just kidding..... */
	int         toggle = 0;

	for (ih = 1; ih <= resolution; ih++) {
		toggle ^= 2;
		for (iv = 1; iv <= resolution; iv++) {
			toggle ^= 1;
			if (sq_wireframe)
				glColor3f(curmat[toggle][0], curmat[toggle][1], curmat[toggle][2]);
			else
				glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, curmat[toggle]);

			xx = cn[iv] * ce[ih];
			zz = cn[iv] * se[ih];
			yy = sn[iv];

			if (sq_wireframe) {
				if ((ih > 1) || (iv > 1)) {
					glBegin(GL_LINES);
					if (ih > 1) {
						glVertex3f(xx, yy, zz);
						glVertex3f(Prevxx[iv], Prevyy[iv], Prevzz[iv]);
					}
					if (iv > 1) {
						glVertex3f(xx, yy, zz);
						glVertex3f(Prevxx[iv - 1], Prevyy[iv - 1], Prevzz[iv - 1]);
					}
/* PURIFY 4.0.1 reports an unitialized memory read on the next line when using
   * MesaGL 2.2 and -sq_mono.  This has been fixed in MesaGL 2.3 and later. */
					glEnd();
				}
			} else {
				if ((cs[iv] > 1e+10) || (cs[iv] < -1e+10)) {
					xn = cs[iv];
					zn = cs[iv];
					yn = ss[iv];
				} else {
					xn = cs[iv] * cw[ih];
					zn = cs[iv] * sw[ih];
					yn = ss[iv];
				}
				if ((ih > 1) && (iv > 1)) {
					glNormal3f(xn, yn, zn);
					glBegin(GL_POLYGON);
					glVertex3f(xx, yy, zz);
					if (!sq_flatshade)
						glNormal3f(Prevxn[iv], Prevyn[iv], Prevzn[iv]);
					glVertex3f(Prevxx[iv], Prevyy[iv], Prevzz[iv]);
					if (!sq_flatshade)
						glNormal3f(xnp, ynp, znp);
					glVertex3f(xp, yp, zp);
					if (!sq_flatshade)
						glNormal3f(Prevxn[iv - 1], Prevyn[iv - 1], Prevzn[iv - 1]);
					glVertex3f(Prevxx[iv - 1], Prevyy[iv - 1], Prevzz[iv - 1]);
					glEnd();
				}
				if (shownorms) {
					if (!sq_flatshade)
						glShadeModel(GL_FLAT);
					glDisable(GL_LIGHTING);
					glBegin(GL_LINES);
					glVertex3f(xx, yy, zz);
					glVertex3f(xx + xn, yy + yn, zz + zn);
					glEnd();
					if (!sq_flatshade)
						glShadeModel(GL_SMOOTH);
					glEnable(GL_LIGHTING);
				}
				xnp = Prevxn[iv];
				ynp = Prevyn[iv];
				znp = Prevzn[iv];
				Prevxn[iv] = xn;
				Prevyn[iv] = yn;
				Prevzn[iv] = zn;
			}

			xp = Prevxx[iv];
			yp = Prevyy[iv];
			zp = Prevzz[iv];
			Prevxx[iv] = xx;
			Prevyy[iv] = yy;
			Prevzz[iv] = zz;

		}		/* next */
	}			/* next */
}

/**** End of really old code ****/

static void
SetCull(int init)
{
	static int  cullmode;

	if (init) {
		cullmode = 0;
		return;
	}
	if (Mode < 1.0001) {
		if (cullmode != 1) {
			glEnable(GL_CULL_FACE);
			glCullFace(GL_BACK);
			cullmode = 1;
		}
	} else if (Mode > 2.9999) {
		if (cullmode != 2) {
			glEnable(GL_CULL_FACE);
			glCullFace(GL_FRONT);
			cullmode = 2;
		}
	} else {
		if (cullmode) {
			glDisable(GL_CULL_FACE);
			cullmode = 0;
		}
	}
}

static void
SetCurrentShape()
{
	int         t;

	xExponent = now.xExponent = later.xExponent;
	yExponent = now.yExponent = later.yExponent;

	for (t = 0; t < 4; ++t) {
		curmat[t][0] = now.r[t] = later.r[t];
		curmat[t][1] = now.g[t] = later.g[t];
		curmat[t][2] = now.b[t] = later.b[t];
	}

	Mode = (double) (now.Mode = later.Mode);
	sq_rotx = now.sq_rotx = later.sq_rotx;
	sq_rotz = now.sq_rotz = later.sq_rotz;

	sq_counter = -maxwait;

	inputs();
}

static void
NextSuperquadric()
{
	double      fnow, flater;
	int         t;

	sq_roty -= spinspeed;
	while (sq_roty >= 360.0)
		sq_roty -= 360.0;
	while (sq_roty < 0.0)
		sq_roty += 360.0;

	--viewcount;

	if (sq_counter > 0) {
		if (--sq_counter == 0) {
			SetCurrentShape();
			if (sq_counter == 0) {		/* Happens if maxwait == 0 */
				MakeUpStuff(0);
				sq_counter = maxcount;
			}
		} else {
			fnow = (double) sq_counter / (double) maxcount;
			flater = (double) (maxcount - sq_counter) / (double) maxcount;
			xExponent = now.xExponent * fnow + later.xExponent * flater;
			yExponent = now.yExponent * fnow + later.yExponent * flater;

			for (t = 0; t < 4; ++t) {
				curmat[t][0] = now.r[t] * fnow + later.r[t] * flater;
				curmat[t][1] = now.g[t] * fnow + later.g[t] * flater;
				curmat[t][2] = now.b[t] * fnow + later.b[t] * flater;
			}

			Mode = (double) now.Mode * fnow + (double) later.Mode * flater;
			sq_rotx = (double) now.sq_rotx * fnow + (double) later.sq_rotx * flater;
			sq_rotz = (double) now.sq_rotz * fnow + (double) later.sq_rotz * flater;

			inputs();
		}
	} else {
		if (++sq_counter >= 0) {
			MakeUpStuff(0);
			sq_counter = maxcount;
		}
	}
}

static void
DisplaySuperquadrics()
{
	glDrawBuffer(GL_BACK);
	if (sq_wireframe)
		glClear(GL_COLOR_BUFFER_BIT);
	else
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	if (viewcount < 1) {
		viewcount = viewwait;
		reshape_superquadrics(-1, -1);
	}
	glPushMatrix();
	glTranslatef(0.0, 0.0, -((GLfloat) (sq_dist) / 16.0) - (Mode * 3.0 - 1.0));	/* viewing transform  */
	glRotatef(sq_rotx, 1.0, 0.0, 0.0);	/* pitch */
	glRotatef(sq_rotz, 0.0, 0.0, 1.0);	/* bank */
	glRotatef(sq_roty, 0.0, 1.0, 0.0);	/* "spin", like heading but comes after P & B */

	SetCull(0);

	DoneScale();

	glPopMatrix();

	/* Remember to flush & swap the buffers after calling this function! */
}

static void
NextSuperquadricDisplay()
{
	NextSuperquadric();
	DisplaySuperquadrics();
}

#define MINSIZE 200
static void
reshape_superquadrics(int w, int h)
{
	static int  last_w = 0, last_h = 0;
	int         maxsize, cursize;

	if (w < 0) {
		w = last_w;
		h = last_h;
	} else {
		last_w = w;
		last_h = h;
	}
	maxsize = (w < h) ? w : h;
	if (maxsize <= MINSIZE) {
		cursize = maxsize;
	} else {
		cursize = myrand(maxsize - MINSIZE) + MINSIZE;
	}
	if ((w > cursize) && (h > cursize)) {
		glViewport(myrand(w - cursize), myrand(h - cursize), cursize, cursize);
		w = h = cursize;
	} else {
		glViewport(0, 0, w, h);
	}
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(30.0, (GLfloat) w / (GLfloat) h, 0.1, 200.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

static void
InitSuperquadrics(int wfmode, int snorm, int res, int count, float speed)
{
	GLfloat     ambient[] =
	{0.4, 0.4, 0.4, 1.0};
	GLfloat     position[] =
	{10.0, 1.0, 1.0, 10.0};
	GLfloat     mat_diffuse[] =
	{1.0, 0.5, 0.5, 1.0};
	GLfloat     mat_specular[] =
	{0.8, 0.8, 0.8, 1.0};
	GLfloat     mat_shininess[] =
	{50.0};

	int         t;

	for (t = 0; t < 4; ++t)
		curmat[t][3] = 1.0;

	sq_rotx = 35.0;
	sq_roty = 0.0;
	sq_rotz = 0.0;
    sq_dist = 140; //(16 << 3);
	sq_wireframe = sq_flatshade = shownorms = 0;
	maxcount = count;
	if (maxcount < 1)
		maxcount = 1;
	maxwait = maxcount >> 1;
	SetCull(1);

	spinspeed = speed;
	viewcount = viewwait = (maxcount < 2) ? 1 : (maxcount << 3);

	if (res < MinRes)
		res = MinRes;
	if (res > MaxRes)
		res = MaxRes;
	resolution = res;

	if (wfmode == 2)
		sq_flatshade = 1;
	else if (wfmode)
		sq_wireframe = 1;

	if (snorm)
		shownorms = 1;

	if (sq_wireframe) {
		glShadeModel(GL_FLAT);
		glDisable(GL_LIGHTING);
		glColor3f(mat_diffuse[0], mat_diffuse[1], mat_diffuse[2]);
	} else {
		if (sq_flatshade) {
			glShadeModel(GL_FLAT);
			position[0] = 1.0;
			position[3] = 0.0;
		}
		glEnable(GL_LIGHTING);
		glEnable(GL_LIGHT0);
		glDepthFunc(GL_LEQUAL);
		glEnable(GL_DEPTH_TEST);

		glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
		glLightfv(GL_LIGHT0, GL_POSITION, position);

		/*glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, mat_diffuse); */
		glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular);
		glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, mat_shininess);

		glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);

		glFrontFace(GL_CW);
		glEnable(GL_NORMALIZE);
	}

	MakeUpStuff(1);
	SetCurrentShape();
	MakeUpStuff(1);	/* Initialize it */
	sq_counter = maxcount;
}

/* End of superquadrics main functions */

void
init_superquadrics()
{
    sq_mono = 0;

    //InitSuperquadrics(MI_IS_WIREFRAME(mi), 0,
    //		  MI_COUNT(mi), MI_CYCLES(mi), spinspeed);
    //ReshapeSuperquadrics(MI_WIDTH(mi), MI_HEIGHT(mi));
    InitSuperquadrics(sq_wireframe,0,25,40,spinspeed);
    DisplaySuperquadrics();
    glFinish();
}

void
draw_superquadrics()
{

	NextSuperquadricDisplay();

	glFinish();
}

void
refresh_superquadrics()
{
	/* Nothing happens here */
}

void
release_superquadrics()
{
}
